from pathlib import Path

import pycodestyle


def test_style_frontend():
    s = pycodestyle.StyleGuide()
    p = (Path(__file__).parents[1].resolve()) / "nmrpoise"
    res = s.check_files([f for f in p.iterdir() if f.suffix == ".py"])
    assert res.total_errors == 0


def test_style_backend():
    s = pycodestyle.StyleGuide()
    p = (Path(__file__).parents[1].resolve()) / "nmrpoise" / "poise_backend"
    res = s.check_files([f for f in p.iterdir() if f.suffix == ".py"])
    assert res.total_errors == 0


def test_style_tests():   # how meta
    s = pycodestyle.StyleGuide()
    p = Path(__file__).parent
    res = s.check_files([f for f in p.iterdir() if f.suffix == ".py"])
    assert res.total_errors == 0



import platform
import subprocess
from pathlib import Path
from distutils.dir_util import copy_tree, remove_tree
from distutils.file_util import copy_file

import pytest

hostname = platform.node()
pkg_toplevel_dir = Path(__file__).parents[1].resolve()

# Have to hardcode these. I don't see a way around it.
hostname_to_tspath = {
    "Empoleon": Path("/opt/topspin4.0.9/exp/stan/nmr"),
    "CARP-CRL": Path(r"C:\Bruker\TopSpin4.0.7\exp\stan\nmr"),
}


def delete_file_force(path):
    """Equivalent to path.unlink(missing_ok=True), but that parameter is
    only in Python 3.8."""
    try:
        path.unlink()
    except FileNotFoundError:
        pass


@pytest.mark.skipif(all(host not in platform.node()
                        for host in hostname_to_tspath.keys()),
                    reason=f"Testing on {hostname} is not supported.")
def test_topspin_installation(tmpdir):
    # Path to TopSpin py/user, and /au/src/user.
    for host in hostname_to_tspath.keys():
        if host in platform.node():
            py_user_path = hostname_to_tspath[host] / "py" / "user"
            au_src_user_path = hostname_to_tspath[host] / "au" / "src" / "user"
    # Make sure it's right
    assert py_user_path.is_dir()

    # Clear out the existing installation of poise
    delete_file_force(py_user_path / "poise.py")
    remove_tree(str(py_user_path / "poise_backend"))
    delete_file_force(au_src_user_path / "poise_1d")
    delete_file_force(au_src_user_path / "poise_2d")
    assert not (py_user_path / "poise.py").exists()
    assert not (py_user_path / "poise_backend").exists()
    assert not (au_src_user_path / "poise_1d").exists()
    assert not (au_src_user_path / "poise_2d").exists()

    # Copy package to a temporary directory
    tmpdir = Path(tmpdir)
    tmpdir_poise = tmpdir / "nmrpoise"
    poise_folder = pkg_toplevel_dir / "nmrpoise"
    ts_install_script = poise_folder / "topspin_install.py"
    copy_file(str(ts_install_script), str(tmpdir))
    copy_tree(str(poise_folder), str(tmpdir_poise))

    # Run the installation script in the temporary directory
    new_ts_install_script = tmpdir_poise / "topspin_install.py"
    subprocess.run(["python3", str(new_ts_install_script)])

    # Check whether the files were installed to TopSpin correctly
    assert (py_user_path / "poise.py").exists()
    assert (py_user_path / "poise_backend").exists()
    assert (py_user_path / "poise_backend" / "backend.py").exists()
    assert (py_user_path / "poise_backend" / "optpoise.py").exists()
    assert (py_user_path / "poise_backend" / "costfunctions.py").exists()
    assert (py_user_path / "poise_backend" / "cfhelpers.py").exists()
    assert (py_user_path / "poise_backend" / "routines").exists()
    assert (au_src_user_path / "poise_1d").exists()
    assert (au_src_user_path / "poise_2d").exists()



import numpy as np
import pytest

from nmrpoise.poise_backend.optpoise import (nelder_mead,
                                             multid_search,
                                             pybobyqa_interface,
                                             deco_count,
                                             scale,
                                             unscale)


def test_scale_unscale():
    # Scale by bounds.
    rng = np.random.default_rng()
    # Generate and sort a random 3x20 matrix.
    mat = rng.uniform(low=-500, high=500, size=(3, 20))
    mat.sort(axis=0)
    lb, val, ub = mat[0], mat[1], mat[2]
    # Generate random percentage tolerances.
    tol = rng.uniform(low=0.01, high=0.5, size=(1, 20))
    # Convert to actual tolerances.
    tol = tol * (ub - lb)
    # Scale.
    sval, slb, sub, stol = scale(val, lb, ub, tol)
    # Make sure that lb's and ub's are zeros and ones.
    assert np.allclose(slb, np.zeros(np.shape(val)))
    assert np.allclose(sub, np.ones(np.shape(val)))
    # Make sure that unscaling gives the right things.
    usval = unscale(sval, lb, ub, tol)
    assert np.allclose(usval, val)

    # Scale by tols.
    # Generate and sort a random 3x20 matrix.
    mat = rng.uniform(low=-500, high=500, size=(3, 20))
    mat.sort(axis=0)
    lb, val, ub = mat[0], mat[1], mat[2]
    # Generate random percentage tolerances.
    tol = rng.uniform(low=0.01, high=0.5, size=(1, 20))
    # Convert to actual tolerances.
    tol = tol * (ub - lb)
    # Scale by tolerances.
    sval, slb, sub, stol = scale(val, lb, ub, tol, scaleby="tols")
    # Make sure that lb's and tol's are 0's and 0.03's.
    assert np.allclose(slb, np.zeros(np.shape(val)))
    assert np.allclose(stol, 0.03 * np.ones(np.shape(val)))
    # Check that unscaling gives back the same values.
    usval = unscale(sval, lb, ub, tol, scaleby="tols")
    assert np.allclose(usval, val)

    # Check other scaleby values for scale and unscale
    with pytest.raises(ValueError, match="Invalid argument"):
        _, _, _, _ = scale(val, lb, ub, tol, scaleby="penguins")
    with pytest.raises(ValueError, match="Invalid argument"):
        _ = unscale(sval, lb, ub, tol, scaleby="penguins")

    # Check out of bounds values
    # Generate and sort a random 3x20 matrix.
    mat = rng.uniform(low=-500, high=500, size=(3, 20))
    mat.sort(axis=0)
    lb, val, ub = mat[1], mat[0], mat[2]  # note how this is out of order
    assert scale(val, lb, ub, tol, scaleby="bounds") is None
    lb, val, ub = mat[0], mat[2], mat[1]  # note how this is out of order
    assert scale(val, lb, ub, tol, scaleby="bounds") is None


def test_deco_count():
    @deco_count
    def peep():
        print("PEEP")

    n = 10
    for i in range(n):
        peep()
    assert peep.calls == n


@deco_count
def rosenbrock(x, arg1=None, arg2=None):
    # Using scipy's definition. We have a couple of dummy arguments because
    # because PyBOBYQA *requires* optimargs and then passes them to the cost
    # function, so rosenbrock() has to accept a couple of extra arguments,
    # even though it doesn't do anything with them.
    return sum(100.0*(x[1:] - x[:-1]**2.0)**2.0 + (1 - x[:-1])**2.0)


@deco_count
def quadratic(x):
    return np.sum(x ** 2)


# This xtol is extremely large, but more closely parallels the
# real-life situations we are likely to face.
lb = np.array([-5, -5, -5, -5, -5])
ub = np.array([5, 5, 5, 5, 5])
x0 = np.array([1.3, 0.7, 0.8, 1.9, 1.2])
xtol = np.array([1e-2] * len(x0))
nm_simplex_methods_nexpts = {
    "spendley": 1,
    "axis": 1,
    "random": 20,
}


def test_errors():
    with pytest.raises(ValueError):
        optResult = nelder_mead(cf=quadratic, x0=x0, xtol=xtol,
                                scaled_lb=lb, scaled_ub=ub,
                                simplex_method="Penguin")
    with pytest.raises(ValueError):
        optResult = nelder_mead(cf=quadratic, x0=x0[:-1], xtol=xtol,
                                scaled_lb=lb, scaled_ub=ub,
                                simplex_method="Penguin")
    with pytest.raises(ValueError):
        optResult = nelder_mead(cf=quadratic, x0=x0, xtol=np.ones(len(x0)),
                                scaled_lb=lb, scaled_ub=ub,
                                simplex_method="Penguin")


def test_NM_accuracy():
    for method, nexpt in nm_simplex_methods_nexpts.items():
        quadratic.calls = 0  # reset fevals
        optResult = nelder_mead(cf=quadratic, x0=x0, xtol=xtol,
                                scaled_lb=lb, scaled_ub=ub,
                                simplex_method=method)
        if not np.allclose(optResult.xbest, np.zeros(len(x0)), atol=2e-2):
            quadratic.calls = 0  # reset fevals
            optResult = nelder_mead(cf=quadratic, x0=x0, xtol=xtol,
                                    scaled_lb=lb, scaled_ub=ub,
                                    simplex_method=method)
        assert np.allclose(optResult.xbest, np.zeros(len(x0)), atol=2e-2)


def test_NM_iters():
    for method, nexpt in nm_simplex_methods_nexpts.items():
        count = 0
        for i in range(nexpt):
            quadratic.calls = 0  # reset fevals
            optResult = nelder_mead(cf=quadratic, x0=x0, xtol=xtol,
                                    scaled_lb=lb, scaled_ub=ub,
                                    simplex_method=method)
            # Check optResult.__repr__() while we're at it.
            assert all(s in repr(optResult) for s in ["xbest", "fbest",
                                                      "niter", "nfev",
                                                      "fvals", "simplex"])
            count += optResult.niter
        count /= nexpt
        assert count < 500


def test_NM_fevals():
    for method, nexpt in nm_simplex_methods_nexpts.items():
        count = 0
        for i in range(nexpt):
            quadratic.calls = 0  # reset fevals
            optResult = nelder_mead(cf=quadratic, x0=x0, xtol=xtol,
                                    scaled_lb=lb, scaled_ub=ub,
                                    simplex_method=method)
            count += optResult.nfev
        count /= nexpt
        assert count < 850


mds_simplex_methods_nexpts = {
    "spendley": 1,
    "axis": 1,
}


def test_MDS_accuracy():
    for method, nexpt in mds_simplex_methods_nexpts.items():
        quadratic.calls = 0  # reset fevals
        optResult = multid_search(cf=quadratic, x0=x0, xtol=xtol,
                                  scaled_lb=lb, scaled_ub=ub,
                                  simplex_method=method)
        assert np.allclose(optResult.xbest, np.zeros(len(x0)), atol=2e-2)


def test_MDS_iters():
    for method, nexpt in mds_simplex_methods_nexpts.items():
        count = 0
        for i in range(nexpt):
            quadratic.calls = 0  # reset fevals
            optResult = multid_search(cf=quadratic, x0=x0, xtol=xtol,
                                      scaled_lb=lb, scaled_ub=ub,
                                      simplex_method=method)
            count += optResult.niter
        count /= nexpt
        assert count < 500


def test_MDS_fevals():
    for method, nexpt in mds_simplex_methods_nexpts.items():
        count = 0
        for i in range(nexpt):
            quadratic.calls = 0  # reset fevals
            optResult = multid_search(cf=quadratic, x0=x0, xtol=xtol,
                                      scaled_lb=lb, scaled_ub=ub,
                                      simplex_method=method)
            count += optResult.nfev
        count /= nexpt
        assert count < 1000


def test_bobyqa_accuracy():
    quadratic.calls = 0  # reset fevals
    sval, slb, sub, stol = scale(x0, lb, ub, xtol, scaleby="tols")
    sbest, _, _, _ = scale(np.zeros(len(x0)), lb, ub, xtol, scaleby="tols")

    def scaled_quadratic(x):
        return np.sum((x - sbest) ** 2)

    optResult = pybobyqa_interface(cf=scaled_quadratic, x0=sval, xtol=stol,
                                   scaled_lb=slb, scaled_ub=sub)
    unscaled_xbest = unscale(optResult.xbest, lb, ub, xtol, scaleby="tols")
    assert np.allclose(unscaled_xbest, np.zeros(len(x0)), atol=2e-2)


def test_maxfevals_reached():
    optResult = nelder_mead(cf=quadratic, x0=x0, xtol=([1e-6] * len(x0)),
                            scaled_lb=lb, scaled_ub=ub, maxfev=10)
    assert optResult.message == "Maximum function evaluations reached."
    optResult = multid_search(cf=quadratic, x0=x0, xtol=([1e-6] * len(x0)),
                              scaled_lb=lb, scaled_ub=ub, maxfev=10)
    assert optResult.message == "Maximum function evaluations reached."



import os
from pathlib import Path

import numpy as np

from nmrpoise.poise_backend import backend as be


def test_get_routine_cf():
    routine_id = "p1cal"   # serialised on AV600
    p_routine_dir = Path(__file__).parent / "test_data"

    routine, cf = be.get_routine_cf(routine_id, p_routine_dir)
    assert routine.name == "p1cal"
    assert routine.pars == ["p1"]
    assert routine.lb == [40]
    assert routine.ub == [56]
    assert routine.init == [48]
    assert routine.tol == [0.2]
    assert routine.cf == "minabsint"
    assert routine.au == "poise_1d"


def test_pidfile():
    pid = os.getpid()
    # We need to be careful with this. pidfile() creates itself in the
    # same directory as backend.py, wherever it is. Since tox testing uses
    # the local copy, we need to check the poise_backend directory.
    pid_fname = (Path(__file__).parents[1] / "nmrpoise" / "poise_backend"
                 / (".pid" + str(pid)))
    # Check that the file exists inside the context manager block.
    with be.pidfile() as _:
        assert pid_fname.exists()
    # Check that it is deleted.
    assert not pid_fname.exists()
